// ----------------------------------------------------------------------------
// $Id: Interpreter.java,v 1.3 1999/07/12 13:07:16 schreine Exp $
// the parser controlling the interpretation
//
// (c) 1999, Wolfgang Schreiner, see file COPYRIGHT
// http://www.risc.uni-linz.ac.at/software/formal
// ----------------------------------------------------------------------------
package Logic;

import Logic.Formulas.*;
import Logic.Terms.*;
import java.io.*;
import java.util.*;
import java.awt.*;

public final class Interpreter
{
  // --------------------------------------------------------------------------
  // scanner tokens
  // --------------------------------------------------------------------------

  private final static int LPAR = '(';      // 40
  private final static int RPAR = ')';      // 41
  private final static int COMMA = ',';     // 44
  private final static int COLON = ':';     // 58
  private final static int SEMICOLON = ';'; // 59
  
  private final static String FUN = "fun";
  private final static String PRED = "pred";
  private final static String RECURSIVE = "recursive";
  private final static String FORMULA = "formula";
  private final static String TERM = "term";
  private final static String READ = "read";
  private final static String OPTION = "option";
  private final static String HELP = "help";
  private final static String PLOT = "plot";
  private final static String PLOTS = "plots";

  private final static String FORALL = "forall";
  private final static String FORALL1 = "forall1";
  private final static String EXISTS = "exists";
  private final static String EXISTS1 = "exists1";
  private final static String AND = "and";
  private final static String AND2 = "and2";
  private final static String OR = "or";
  private final static String OR2 = "or2";
  private final static String NOT = "not";
  private final static String IMPLIES = "implies";
  private final static String EQUIV = "equiv";
  private final static String LET = "let";
  private final static String LET1 = "let1";
  private final static String IF = "if";
  private final static String SUCH = "such";

  private final static String SET = "set";
  private final static String SET1 = "set1";
  private final static String REDUCE = "reduce";
  private final static String TUPLE = "tuple";

  private final static String FDEF = "=";
  private final static String PDEF = "<=>";
  private final static String IN = "in";

  // --------------------------------------------------------------------------
  // private variables
  // --------------------------------------------------------------------------

  // set if definition is parsed (check for undefined constants/variables)
  private boolean parseDef = false;

  // set if error occurred during parsing of definition
  private boolean errorDef = false;

  // control object
  private Control control;

  // --------------------------------------------------------------------------
  // create interpreter with denoted control object
  // --------------------------------------------------------------------------
  public Interpreter(Control _control)
  {
    control = _control;
  }

  // --------------------------------------------------------------------------
  // parse input on stream 'in' and execute
  // --------------------------------------------------------------------------
  public void run(Reader reader)
  {
    // stream tokenizer does not handle // comments correctly
    reader = new CommentFilter(new BufferedReader(reader));
    StreamTokenizer st = new StreamTokenizer(reader);

    st.resetSyntax();
    st.whitespaceChars(0, 32);
    st.wordChars(33, 39);
    st.wordChars(42, 43);
    st.wordChars(45, 57);
    st.wordChars(60, 126);

    try
      {
	int token = st.nextToken();
	while (token != st.TT_EOF)
	  {
	    boolean okay = false;
	    if (token == st.TT_WORD)
	      {
		if (st.sval.equals(PRED))
		  {
		    parseDef = true;
		    errorDef = false;
		    Predicate pred = parsePredicateDefinition(st);
		    parseDef = false;
		    okay = pred != null;
		    if (errorDef)
		      control.println("> definition is ignored.");
		    else
		      if (okay) definePredicate(pred);
		  }
		else if (st.sval.equals(FUN))
		  {
		    parseDef = true;
		    errorDef = false;
		    Function fun = parseFunctionDefinition(st);
		    parseDef = false;
		    okay = fun != null;
		    if (errorDef)
		      control.println("> definition is ignored.");
		    else 
		      if (okay) defineFunction(fun);
		  }
		else if (st.sval.equals(FORMULA))
		  {
		    Formula formula = parseFormula(st);
		    okay = (formula != null);
		    if (okay) evalFormula(formula);
		  }
		else if (st.sval.equals(TERM))
		  {
		    Term term = parseTerm(st);
		    okay = (term != null);
		    if (okay) evalTerm(term);
		  }
		else if (st.sval.equals(READ))
		  {
		    String string = parseString(st);
		    okay = (string != null);
		    if (okay) readInput(string);
		  }
		else if (st.sval.equals(OPTION))
		  {		    
		    okay = setOption(st);
		  }
		else if (st.sval.equals(HELP))
		{
		  help();
		  okay = true;
		}
		else if (st.sval.equals(PLOT))
		{
		  Term term = parseTerm(st);
		  okay = (term != null);
		  if (okay) plotRelation(term);
		}
		else if (st.sval.equals(PLOTS))
		{
		  Term term = parseTerm(st);
		  okay = (term != null);
		  if (okay) plotClasses(term);
		}
		else
		  {
		    okay = false;
		  }
	      }
	    if (okay)
	      {
		token = st.nextToken();
		okay = (token == SEMICOLON);
		if (!okay)
		  printError("\";\" expected");
	      }
	    if (!okay)
	      {
		st.pushBack();
		token = st.nextToken();
		if (token == st.TT_WORD)		  
		  printError("unexpected '" + st.sval + "'");
		else if (token == st.TT_EOF)
		  printError("unexpected end of file");
		else
		  {
		    byte[] b = new byte[1];
		    b[0] = (byte)token;
		    printError("unexpected '" + new String(b) +"'");
		  }
		while (token != st.TT_EOF && token != SEMICOLON)
		  token = st.nextToken();
	      }
	    while (token == SEMICOLON)
	      token = st.nextToken();
	  }
      }
    catch(IOException e)
      {
	printError("IOException " + e.getMessage());
	control.abort();
      }
  }

  // --------------------------------------------------------------------------
  // help message
  // --------------------------------------------------------------------------
  private void help()
  {
    control.println("> commands (terminated by ';'):");
    control.println(">   pred <name>(<parameters>) <=> <formula>;");
    control.println(">   fun <name>(<parameters>) = <term>;");
    control.println(">   formula <formula>;");
    control.println(">   term <term>;");
    control.println(">   plot <term>;");
    control.println(">   plots <term>;");
    control.println(">   read <name>;");
    control.println(">   option <name> = <value>;");
    control.println(">   help;");
    control.println("> see " + Main.url + ".");
  }

  // --------------------------------------------------------------------------
  //
  // syntactic functions
  //
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  // name
  // --------------------------------------------------------------------------
  private String parseString(StreamTokenizer st) 
       throws IOException
  {
    int token = st.nextToken();
    if (token != st.TT_WORD) return null;
    return st.sval;
  }

  // --------------------------------------------------------------------------
  // name "in" term
  // --------------------------------------------------------------------------
  private TypedVariable parseTypedVariable(StreamTokenizer st) 
       throws IOException
  {
    String name = parseString(st);
    if (name == null) return null;
    int token = st.nextToken();
    Term domain = parseTerm(st);
    if (domain == null) return null;
    return new TypedVariable(name, domain);
  }

  // --------------------------------------------------------------------------
  // name [ "in" term | "=" term ]
  // --------------------------------------------------------------------------
  private IteratorVariable parseIteratorVariable(StreamTokenizer st) 
    throws IOException
  {
    String name = parseString(st);
    if (name == null) return null;
    int token = st.nextToken();
    String key = st.sval;
    if (token != st.TT_WORD || !(key.equals(IN) || key.equals(FDEF)))
      {
	st.pushBack();
	return new Variable(name);
      }
    Term term = parseTerm(st);
    if (term == null) return null;
    if (key.equals(IN))
      return new TypedVariable(name, term);
    else
      return new ValueVariable(name, term);
  }

  // --------------------------------------------------------------------------
  // { IteratorVariable }
  // --------------------------------------------------------------------------
  private IteratorVariable[] parseIteratorVariables(StreamTokenizer st) 
    throws IOException
  {
    Vector vars = new Vector();
    int token = st.nextToken();
    while (token != RPAR && token != COLON)
      {
	st.pushBack();
	IteratorVariable var = parseIteratorVariable(st);
	if (var == null) return null;
	vars.addElement(var);
	Context.begin(var.name(), new Set());
	token = st.nextToken();
	if (token == COMMA)
	  token = st.nextToken();
	else if (token != COLON)
	  return null;
      }
    st.pushBack();
    int n = vars.size();
    for (int i=0; i<n; i++)
      Context.end();
    IteratorVariable[] vars0 = new IteratorVariable[n];
    vars.copyInto(vars0);
    return vars0;
  }

  // --------------------------------------------------------------------------
  // name [ "in" term | ":" name ]
  // --------------------------------------------------------------------------
  private ConstrainedVariable parseConstrainedVariable(StreamTokenizer st) 
    throws IOException
  {
    String name = parseString(st);
    if (name == null) return null;
    int token = st.nextToken();
    if (token != COLON && !(token == st.TT_WORD && st.sval.equals(IN)))	
      {
	st.pushBack();
	return new Variable(name);
      }
    if (token == COLON)
    {
      String pname = parseString(st);
      if (pname == null) return null;
      Predicate pred = Model.getPredicate(pname, 1);
      if (pred == null)
      {
	errorDef = true;
	printError("undefined predicate " + pname + "/1");
      }
      return new CheckedVariable(name, pname);
    }
    else
    {
      Term domain = parseTerm(st);
      if (domain == null) return null;
      return new TypedVariable(name, domain);
    }
  }

  // --------------------------------------------------------------------------
  // { ConstrainedVariable }
  // --------------------------------------------------------------------------
  private ConstrainedVariable[] parseConstrainedVariables(StreamTokenizer st) 
       throws IOException
  {
    Vector vars = new Vector();
    int token = st.nextToken();
    while (token != RPAR)
      {
	st.pushBack();
	ConstrainedVariable var = parseConstrainedVariable(st);
	if (var == null) return null;
	vars.addElement(var);
	Context.begin(var.name(), new Set());
	token = st.nextToken();
	if (token == COMMA)
	  token = st.nextToken();
	else if (token != RPAR)
	  return null;
      }
    st.pushBack();
    int n = vars.size();
    for (int i=0; i<n; i++)
      Context.end();
    ConstrainedVariable[] vars0 = new ConstrainedVariable[n];
    vars.copyInto(vars0);
    return vars0;
  }

  // --------------------------------------------------------------------------
  // name "=" term 
  // --------------------------------------------------------------------------
  private BoundVariable parseBoundVariable(StreamTokenizer st) 
       throws IOException
  {
    String name = parseString(st);
    if (name == null) return null;
    String fdef = parseString(st);
    if (fdef == null) return null;
    if (!fdef.equals(FDEF)) return null;
    Term value = parseTerm(st);
    if (value == null) return null;
    return new BoundVariable(name, value);
  }

  // --------------------------------------------------------------------------
  // { BoundVariable }
  // --------------------------------------------------------------------------
  private BoundVariable[] parseBoundVariables(StreamTokenizer st) 
       throws IOException
  {
    Vector vars = new Vector();
    int token = st.nextToken();
    while (token != COLON)
      {
	st.pushBack();
	BoundVariable var = parseBoundVariable(st);
	if (var == null) return null;
	vars.addElement(var);
	Context.begin(var.name(), new Set());
	token = st.nextToken();
	if (token == COMMA)
	  token = st.nextToken();
	else if (token != COLON)
	  return null;
      }
    st.pushBack();
    int n = vars.size();
    for (int i=0; i<n; i++)
      Context.end();
    BoundVariable[] vars0 = new BoundVariable[vars.size()];
    vars.copyInto(vars0);
    return vars0;
  }

  // --------------------------------------------------------------------------
  // [ "(" ")" ]
  // --------------------------------------------------------------------------
  private boolean parseParentheses(StreamTokenizer st) throws IOException
  {
    int token = st.nextToken();
    if (token == LPAR)
      {
	token = st.nextToken();
	if (token != RPAR) 
	  {
	    st.pushBack();
	    return false;
	  }
      }
    else
      st.pushBack();
    return true;
  }

  // --------------------------------------------------------------------------
  // term sequences
  // --------------------------------------------------------------------------
  private Term[] parseTerms(StreamTokenizer st) throws IOException
  {
    Vector terms = new Vector();
    int token = st.nextToken();
    while (token != RPAR)
      {
	st.pushBack();
	Term term = parseTerm(st);
	if (term == null) return null;
	terms.addElement(term);
	token = st.nextToken();
	if (token == COMMA)
	  token = st.nextToken();
	else if (token != RPAR)
	  return null;
      }
    st.pushBack();
    Term[] terms0 = new Term[terms.size()];
    terms.copyInto(terms0);
    return terms0;
  }

  // --------------------------------------------------------------------------
  // formula sequences
  // --------------------------------------------------------------------------
  private Formula[] parseFormulas(StreamTokenizer st) throws IOException
  {
    Vector formulas = new Vector();
    int token = st.nextToken();
    while (token != RPAR)
      {
	st.pushBack();
	Formula formula = parseFormula(st);
	if (formula == null) return null;
	formulas.addElement(formula);
	token = st.nextToken();
	if (token == COMMA)
	  token = st.nextToken();
	else if (token != RPAR)
	  return null;
      }
    st.pushBack();
    Formula[] formulas0 = new Formula[formulas.size()];
    formulas.copyInto(formulas0);
    return formulas0;
  }

  // --------------------------------------------------------------------------
  // formulas
  // --------------------------------------------------------------------------
  private Formula parseFormula(StreamTokenizer st) throws IOException
  {
    int token = st.nextToken();
    if (token != st.TT_WORD) return null;
    // ------------------------------------------------------------------------
    // not "(" formula ")"
    // ------------------------------------------------------------------------
    if (st.sval.equals(NOT))
      {
	token = st.nextToken();
	if (token != LPAR) return null;
	Formula f = parseFormula(st);
	if (f == null) return null;
	token = st.nextToken();
	if (token != RPAR) return null;
	return new Not(f);
      }
    // ------------------------------------------------------------------------
    // and2 "(" formula "," formula ")"
    // ------------------------------------------------------------------------
    else if (st.sval.equals(AND2))
      {
	token = st.nextToken();
	if (token != LPAR) return null;
	Formula f0 = parseFormula(st);
	if (f0 == null) return null;
	token = st.nextToken();
	if (token != COMMA) return null;
	Formula f1 = parseFormula(st);
	if (f1 == null) return null;
	token = st.nextToken();
	if (token != RPAR) return null;
	return new And(f0, f1);
      }
    // ------------------------------------------------------------------------
    // and "(" { formula } ")"
    // ------------------------------------------------------------------------
    else if (st.sval.equals(AND))
      {
	token = st.nextToken();
	if (token != LPAR) return null;
	Formula[] f = parseFormulas(st);
	if (f == null) return null;
	token = st.nextToken();
	if (token != RPAR) return null;
	return new AndN(f);
      }
    // ------------------------------------------------------------------------
    // or2 "(" formula "," formula ")"
    // ------------------------------------------------------------------------
    else if (st.sval.equals(OR2))
      {
	token = st.nextToken();
	if (token != LPAR) return null;
	Formula f0 = parseFormula(st);
	if (f0 == null) return null;
	token = st.nextToken();
	if (token != COMMA) return null;
	Formula f1 = parseFormula(st);
	if (f1 == null) return null;
	token = st.nextToken();
	if (token != RPAR) return null;
	return new Or(f0, f1);
      }
    // ------------------------------------------------------------------------
    // or "(" { formula } ")"
    // ------------------------------------------------------------------------
    else if (st.sval.equals(OR))
      {
	token = st.nextToken();
	if (token != LPAR) return null;
	Formula[] f = parseFormulas(st);
	if (f == null) return null;
	token = st.nextToken();
	if (token != RPAR) return null;
	return new OrN(f);
      }
    // ------------------------------------------------------------------------
    // implies "(" formula "," formula ")"
    // ------------------------------------------------------------------------
    else if (st.sval.equals(IMPLIES))
      {
	token = st.nextToken();
	if (token != LPAR) return null;
	Formula f0 = parseFormula(st);
	if (f0 == null) return null;
	token = st.nextToken();
	if (token != COMMA) return null;
	Formula f1 = parseFormula(st);
	if (f1 == null) return null;
	token = st.nextToken();
	if (token != RPAR) return null;
	return new Implies(f0, f1);
      }
    // ------------------------------------------------------------------------
    // equiv "(" formula "," formula ")"
    // ------------------------------------------------------------------------
    else if (st.sval.equals(EQUIV))
      {
	token = st.nextToken();
	if (token != LPAR) return null;
	Formula f0 = parseFormula(st);
	if (f0 == null) return null;
	token = st.nextToken();
	if (token != COMMA) return null;
	Formula f1 = parseFormula(st);
	if (f1 == null) return null;
	token = st.nextToken();
	if (token != RPAR) return null;
	return new Equiv(f0, f1);
      }
    // ------------------------------------------------------------------------
    // forall1 "(" TypedVariable ":" formula ")"
    // ------------------------------------------------------------------------
    else if (st.sval.equals(FORALL1))
      {
	token = st.nextToken();
	if (token != LPAR) return null;
	TypedVariable var = parseTypedVariable(st);
	if (var == null) return null;
	token = st.nextToken();
	if (token != COLON) return null;
	Context.begin(var.name(), new Set());
	Formula formula = parseFormula(st);
	Context.end();
	if (formula == null) return null;
	token = st.nextToken();
	if (token != RPAR) return null;
	return new ForAll(var.name(), var.domain(), formula);
      }
    // ------------------------------------------------------------------------
    // forall "(" { TypedVariable } ":" formula ")"
    // ------------------------------------------------------------------------
    else if (st.sval.equals(FORALL))
      {
	token = st.nextToken();
	if (token != LPAR) return null;
	IteratorVariable[] vars = parseIteratorVariables(st);
	if (vars == null) return null;
	token = st.nextToken();
	if (token != COLON) return null;
	for (int i=0; i<vars.length; i++)
	  Context.begin(vars[i].name(), new Set());
	Formula formula = parseFormula(st);
	for (int i=0; i<vars.length; i++)
	  Context.end();
	if (formula == null) return null;
	token = st.nextToken();
	if (token != RPAR) return null;
	return new ForAllN(vars, formula);
      }
    // ------------------------------------------------------------------------
    // exists1 "(" TypedVariable ":" formula ")"
    // ------------------------------------------------------------------------
    else if (st.sval.equals(EXISTS1))
      {
	token = st.nextToken();
	if (token != LPAR) return null;
	TypedVariable var = parseTypedVariable(st);
	if (var == null) return null;
	token = st.nextToken();
	if (token != COLON) return null;
	Context.begin(var.name(), new Set());
	Formula formula = parseFormula(st);
	Context.end();
	if (formula == null) return null;
	token = st.nextToken();
	if (token != RPAR) return null;
	return new Exists(var.name(), var.domain(), formula);
      }
    // ------------------------------------------------------------------------
    // exists "(" { TypedVariable } ":" formula ")"
    // ------------------------------------------------------------------------
    else if (st.sval.equals(EXISTS))
      {
	token = st.nextToken();
	if (token != LPAR) return null;
	IteratorVariable[] vars = parseIteratorVariables(st);
	if (vars == null) return null;
	token = st.nextToken();
	if (token != COLON) return null;
	for (int i=0; i<vars.length; i++)
	  Context.begin(vars[i].name(), new Set());
	Formula formula = parseFormula(st);
	for (int i=0; i<vars.length; i++)
	  Context.end();
	if (formula == null) return null;
	token = st.nextToken();
	if (token != RPAR) return null;
	return new ExistsN(vars, formula);
      }
    // ------------------------------------------------------------------------
    // let1 "(" name "=" term ":" formula ")" 
    // ------------------------------------------------------------------------
    else if (st.sval.equals(LET1))
      {
	token = st.nextToken();
	if (token != LPAR) return null;
	BoundVariable var = parseBoundVariable(st);
	if (var == null) return null;
	token = st.nextToken();
	if (token != COLON) return null;	
	Context.begin(var.name(), new Set());
	Formula formula = parseFormula(st);
	Context.end();
	if (formula == null) return null;
	token = st.nextToken();
	if (token != RPAR) return null;
	return new LetFormula(var.name(), var.value(), formula);
      }
    // ------------------------------------------------------------------------
    // let "(" { name "=" term } ":" formula ")" 
    // ------------------------------------------------------------------------
    else if (st.sval.equals(LET))
      {
	token = st.nextToken();
	if (token != LPAR) return null;
	BoundVariable[] vars = parseBoundVariables(st);
	if (vars == null) return null;
	token = st.nextToken();
	if (token != COLON) return null;
	for (int i=0; i<vars.length; i++)
	  Context.begin(vars[i].name(), new Set());
	Formula formula = parseFormula(st);
	for (int i=0; i<vars.length; i++)
	  Context.end();
	if (formula == null) return null;
	token = st.nextToken();
	if (token != RPAR) return null;
	return new LetNFormula(vars, formula);
      }
    // ------------------------------------------------------------------------
    // if "(" formula "," formula [ "," formula ")" ]
    // ------------------------------------------------------------------------
    else if (st.sval.equals(IF))
      {
	token = st.nextToken();
	if (token != LPAR) return null;
	Formula guard = parseFormula(st);
	if (guard == null) return null;
	token = st.nextToken();
	if (token != COMMA) return null;
        Formula formula0 = parseFormula(st);
	if (formula0 == null) return null;
	token = st.nextToken();
	if (token != COMMA)
	  {
	    if (token != RPAR) return null;
	    return new IfThenFormula(guard, formula0);
	  }
	Formula formula1 = parseFormula(st);
	if (formula1 == null) return null;
	token = st.nextToken();
	if (token != RPAR) return null;
	return new IfThenElseFormula(guard, formula0, formula1);	    
      }
    // ------------------------------------------------------------------------
    // name [ "(" terms ")" ]
    // ------------------------------------------------------------------------
    else
      {
	String name = st.sval;
	Term[] arguments = null;
	token = st.nextToken();	
	if (token != LPAR)
	  {
	    st.pushBack();
	    arguments = new Term[0];
	  }
	else
	  {
	    arguments = parseTerms(st);
	    if (arguments == null) return null;
	    token = st.nextToken();
	    if (token != RPAR) return null;
	  }
	if (parseDef)
	  {
	    Predicate pred = Model.getPredicate(name, arguments.length);
	    if (pred == null)
	      {
		errorDef = true;
		printError("undefined predicate " + name + "/" +
			   arguments.length);
	      }
	    
	  }
	return new Atomic(name, arguments);
      }
  }

  // --------------------------------------------------------------------------
  // terms
  // --------------------------------------------------------------------------
  private Term parseTerm(StreamTokenizer st) throws IOException
  {
    int token = st.nextToken();
    if (token != st.TT_WORD) return null;
    // ------------------------------------------------------------------------
    // natural number
    // ------------------------------------------------------------------------
    if (Character.isDigit(st.sval.charAt(0)))
      {
	try
	  {
	    Integer value = Integer.valueOf(st.sval);
	    return new NatTerm(value.intValue());
	  }
	catch(NumberFormatException e)
	  {
	    printError("literal '" + st.sval + "' does not denote a number");
	    return null;
	  }
      }
    // ------------------------------------------------------------------------
    // set1 "(" TypedVariable ":" formula "," term ")" 
    // ------------------------------------------------------------------------
    if (st.sval.equals(SET1))
      {
	token = st.nextToken();
	if (token != LPAR) return null;
	TypedVariable var = parseTypedVariable(st);
	if (var == null) return null;
	token = st.nextToken();
	if (token != COLON) return null;
	Context.begin(var.name(), new Set());
	Formula formula = parseFormula(st);
	Context.end();
	if (formula == null) return null;
	token = st.nextToken();
	if (token != COMMA) return null;
	Context.begin(var.name(), new Set());
	Term term = parseTerm(st);
	Context.end();
	if (term == null) return null;
	token = st.nextToken();
	if (token != RPAR) return null;
	return new SetTerm(var.name(), var.domain(), formula, term);
      }
    // ------------------------------------------------------------------------
    // set "(" { IteratorVariable } ":" formula "," term ")" 
    // ------------------------------------------------------------------------
    else if (st.sval.equals(SET))
      {
	token = st.nextToken();
	if (token != LPAR) return null;
	IteratorVariable[] vars = parseIteratorVariables(st);
	if (vars == null) return null;
	token = st.nextToken();
	if (token != COLON) return null;
	for (int i=0; i<vars.length; i++)
	  Context.begin(vars[i].name(), new Set());
	Formula formula = parseFormula(st);
	for (int i=0; i<vars.length; i++)
	  Context.end();
	if (formula == null) return null;
	token = st.nextToken();
	if (token != COMMA) return null;
	for (int i=0; i<vars.length; i++)
	  Context.begin(vars[i].name(), new Set());
	Term term = parseTerm(st);
	for (int i=0; i<vars.length; i++)
	  Context.end();
	if (term == null) return null;
	token = st.nextToken();
	if (token != RPAR) return null;
	return new SetNTerm(vars, formula, term);
      }
    // ------------------------------------------------------------------------
    // reduce "(" name "," term "," term ")"
    // ------------------------------------------------------------------------
    else if (st.sval.equals(REDUCE))
      {
	token = st.nextToken();
	if (token != LPAR) return null;
	String name = parseString(st);
	if (name == null) return null;
	token = st.nextToken();
	if (token != COMMA) return null;
        Term term0 = parseTerm(st);
	if (term0 == null) return null;
	token = st.nextToken();
	if (token != COMMA) return null;
        Term term1 = parseTerm(st);
	if (term1 == null) return null;
	token = st.nextToken();
	if (token != RPAR) return null;
	if (parseDef)
	  {
	    Function fun = Model.getFunction(name, 2);
	    if (fun == null)
	      {
		errorDef = true;
		printError("undefined function " + name + "/2");
	      }
	  }
	return new Reduce(name, term0, term1);
      }
    // ------------------------------------------------------------------------
    // let1 "(" name ":=" term ":" term ")" 
    // ------------------------------------------------------------------------
    else if (st.sval.equals(LET1))
      {
	token = st.nextToken();
	if (token != LPAR) return null;
	BoundVariable var = parseBoundVariable(st);
	if (var == null) return null;
	token = st.nextToken();
	if (token != COLON) return null;	
	Context.begin(var.name(), new Set());
	Term term1 = parseTerm(st);
	Context.end();
	if (term1 == null) return null;
	token = st.nextToken();
	if (token != RPAR) return null;
	return new LetTerm(var.name(), var.value(), term1);
      }
    // ------------------------------------------------------------------------
    // let "(" { BoundVariable } ":" term ")" 
    // ------------------------------------------------------------------------
    else if (st.sval.equals(LET))
      {
	token = st.nextToken();
	if (token != LPAR) return null;
	BoundVariable[] vars = parseBoundVariables(st);
	if (vars == null) return null;
	token = st.nextToken();
	if (token != COLON) return null;
	for (int i=0; i<vars.length; i++)
	  Context.begin(vars[i].name(), new Set());
	Term term = parseTerm(st);
	for (int i=0; i<vars.length; i++)
	  Context.end();
	if (term == null) return null;
	token = st.nextToken();
	if (token != RPAR) return null;
	return new LetNTerm(vars, term);
      }
    // ------------------------------------------------------------------------
    // such "(" { IteratorVariable } ":" formula "," term ")" 
    // ------------------------------------------------------------------------
    else if (st.sval.equals(SUCH))
      {
	token = st.nextToken();
	if (token != LPAR) return null;
	IteratorVariable[] vars = parseIteratorVariables(st);
	if (vars == null) return null;
	token = st.nextToken();
	if (token != COLON) return null;
	for (int i=0; i<vars.length; i++)
	  Context.begin(vars[i].name(), new Set());
	Formula formula = parseFormula(st);
	for (int i=0; i<vars.length; i++)
	  Context.end();
	if (formula == null) return null;
	token = st.nextToken();
	if (token != COMMA) return null;
	for (int i=0; i<vars.length; i++)
	  Context.begin(vars[i].name(), new Set());
	Term term = parseTerm(st);
	for (int i=0; i<vars.length; i++)
	  Context.end();
	if (term == null) return null;
	token = st.nextToken();
	if (token != RPAR) return null;
	return new SuchThat(vars, formula, term);
      }
    // ------------------------------------------------------------------------
    // if "(" formula "," term [ "," term ")" ]
    // ------------------------------------------------------------------------
    else if (st.sval.equals(IF))
      {
	token = st.nextToken();
	if (token != LPAR) return null;
	Formula guard = parseFormula(st);
	if (guard == null) return null;
	token = st.nextToken();
	if (token != COMMA) return null;
        Term term0 = parseTerm(st);
	if (term0 == null) return null;
	token = st.nextToken();
	if (token != COMMA)
	  {
	    if (token != RPAR) return null;
	    return new IfThenTerm(guard, term0);
	  }
	Term term1 = parseTerm(st);
	if (term1 == null) return null;
	token = st.nextToken();
	if (token != RPAR) return null;
	return new IfThenElseTerm(guard, term0, term1);	    
      }
    // ------------------------------------------------------------------------
    // tuple "(" terms ")"
    // ------------------------------------------------------------------------
    else if (st.sval.equals(TUPLE))
      {
	token = st.nextToken();	
	if (token != LPAR) return null;
	Term[] elems = parseTerms(st);
	if (elems == null) return null;
	token = st.nextToken();
	if (token != RPAR) return null;
	return new TupleTerm(elems);
      }
    // ------------------------------------------------------------------------
    // .number "(" term ")"
    // ------------------------------------------------------------------------
    else if (st.sval.charAt(0) == '.')
      {
	int field = 0;
	try
	  {
	    field = Integer.valueOf(st.sval.substring(1)).intValue();	    
	  }
	catch(NumberFormatException e)
	  {
	    printError("literal '" + st.sval.substring(1) + 
		       "' does not denote a number");
	    return null;
	  }
	token = st.nextToken();	
	if (token != LPAR) return null;
	Term tuple = parseTerm(st);
	if (tuple == null) return null;
	token = st.nextToken();
	if (token != RPAR) return null;
	return new Select(tuple, field);
      }
    // ------------------------------------------------------------------------
    // variable
    // name [ "(" terms ")" ]
    // ------------------------------------------------------------------------
    else
      {
	String name = st.sval;
	if (Context.get(name) != null)
	  return new Variable(name);
	Term[] arguments = null;
	token = st.nextToken();
	if (token != LPAR)
	  {
	    st.pushBack();
	    arguments = new Term[0];
	  }
	else
	  {
	    arguments = parseTerms(st);
	    if (arguments == null) return null;
	    token = st.nextToken();
	    if (token != RPAR) return null;
	  }
	if (parseDef)
	  {
	    Function fun = Model.getFunction(name, arguments.length);
	    if (fun == null)
	      {
		errorDef = true;
		printError("undefined function " + name + "/" +
			   arguments.length);
	      }
	  }
	return new Application(name, arguments);
      }	
  }

  // --------------------------------------------------------------------------
  // pred <name> [ "(" <constrained_names> ")" ] "<=>" formula
  // --------------------------------------------------------------------------
  private Predicate parsePredicateDefinition(StreamTokenizer st) 
       throws IOException
  {
    String name = parseString(st);
    if (name == null) return null;
    ConstrainedVariable[] params = null;
    int token = st.nextToken();
    if (token == LPAR) 
      {
	params = parseConstrainedVariables(st);
	if (params == null) return null;
	token = st.nextToken();
	if (token != RPAR) return null;
      }
    else
      {
	st.pushBack();
	params = new ConstrainedVariable[0];
      }
    Predicate old = Model.getPredicate(name, params.length);
    String delim = parseString(st);
    if (delim == null) return null;
    Term recterm = null;
    if (delim.equals(RECURSIVE))
    {
      if (old != null) Model.delete(old);
      for (int j=0; j<params.length; j++)
	Context.begin(params[j].name(), new Set());
      recterm = parseTerm(st);
      for (int j=0; j<params.length; j++)
	Context.end();
      if (old != null) Model.add(old);
      if (recterm == null) return null;
      delim = parseString(st);
    }
    if (!delim.equals(PDEF)) return null;
    if (old != null) Model.delete(old);
    for (int j=0; j<params.length; j++)
      Context.begin(params[j].name(), new Set());
    Predicate dummy = null;
    if (recterm != null)
    {      
      dummy = new DummyPredicate(name, params);
      Model.add(dummy);
    }
    Formula formula = parseFormula(st);
    if (recterm != null)
      Model.delete(dummy);
    for (int j=0; j<params.length; j++)
      Context.end();
    if (old != null) Model.add(old);
    if (formula == null) return null;
    if (recterm == null)
      return new DefinedPredicate(name, params, formula);
    else
      return new RecursivePredicate(name, params, recterm, formula);
  }

  // --------------------------------------------------------------------------
  // fun <name> [ "(" <constrained_names> ")" ] [ recursive term ] "=" term
  // --------------------------------------------------------------------------
  private Function parseFunctionDefinition(StreamTokenizer st) 
       throws IOException
  {
    String name = parseString(st);
    if (name == null) return null;
    ConstrainedVariable[] params = null;
    int token = st.nextToken();
    if (token == LPAR) 
      {
	params = parseConstrainedVariables(st);
	if (params == null) return null;
	token = st.nextToken();
	if (token != RPAR) return null;
      }
    else
      {
	st.pushBack();
	params = new ConstrainedVariable[0];
      }
    Function old = Model.getFunction(name, params.length);
    String delim = parseString(st);
    if (delim == null) return null;
    Term recterm = null;
    if (delim.equals(RECURSIVE))
    {
      if (old != null) Model.delete(old);
      for (int j=0; j<params.length; j++)
	Context.begin(params[j].name(), new Set());
      recterm = parseTerm(st);
      for (int j=0; j<params.length; j++)
	Context.end();
      if (old != null) Model.add(old);
      if (recterm == null) return null;
      delim = parseString(st);
    }
    if (!delim.equals(FDEF)) return null;
    if (old != null) Model.delete(old);
    for (int j=0; j<params.length; j++)
      Context.begin(params[j].name(), new Set());
    Function dummy = null;
    if (recterm != null)
    {
      dummy = new DummyFunction(name, params);
      Model.add(dummy);
    }
    Term term = parseTerm(st);
    if (recterm != null)
      Model.delete(dummy);
    for (int j=0; j<params.length; j++)
      Context.end();
    if (old != null) Model.add(old);
    if (term == null) return null;
    if (recterm == null)
      return new DefinedFunction(name, params, term);
    else
      return new RecursiveFunction(name, params, recterm, term);
  }

  // --------------------------------------------------------------------------
  //
  // semantic functions
  //
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  // define predicate
  // --------------------------------------------------------------------------
  private void definePredicate(Predicate pred) throws IOException
  {
    Predicate pred0 = Model.getPredicate(pred.name(), pred.arity());
    if (pred0 != null)
      Model.delete(pred0);
    Model.add(pred);
    if (!control.isSilent())
      control.println("> predicate " + pred.name() + "/" + pred.arity() + ".");
  }

  // --------------------------------------------------------------------------
  // define function
  // --------------------------------------------------------------------------
  private void defineFunction(Function fun)
  {
    Function fun0 = Model.getFunction(fun.name(), fun.arity());
    if (fun0 != null)
      Model.delete(fun0);
    Model.add(fun);
    if (!control.isSilent())
      control.println("> function " + fun.name() + "/" + fun.arity() + ".");
  }

  // --------------------------------------------------------------------------
  // print error message
  // --------------------------------------------------------------------------
  private void printError(String message)
  {
    int len = message.length();
    control.print("> ERROR: ");
    int start = 0;
    while (true)
      {
	int end = message.indexOf('\n', start);
	if (end == -1)
	  {
	    control.print(message.substring(start));
	    control.println(".");
	    break;
	  }	    
	control.println(message.substring(start, end));
	control.print("> ");
	start = end+1;
	if (start == len) break;
      }
  }

  // --------------------------------------------------------------------------
  // evaluate formula
  // --------------------------------------------------------------------------
  private void evalFormula(Formula formula)
  {
    try
      {
	if (formula.eval())
	  control.println("> true.");
	else
	  control.println("> false.");
      }
    catch(EvalException e)
      {
	// context is in an undefined state
	Context.reset();
	printError(e.getMessage());
      }
  }

  // --------------------------------------------------------------------------
  // evaluate term
  // --------------------------------------------------------------------------
  private void evalTerm(Term term)
  {
    try
      {
	Value value = term.eval();
	control.print("> ");
	control.print(value.getString());
	control.println(".");
      }
    catch(EvalException e)
      {
	// context is in an undefined state
	Context.reset();
	printError(e.getMessage());
      }
  }

  // --------------------------------------------------------------------------
  // read input from file denoted by `path`
  // --------------------------------------------------------------------------
  private void readInput(String path)
  {
    // netscape only allows resources with certain extensions    
    path = path + ".txt";
    try
      {
	InputStream input = null;
	if (control.isApplet())
	  input = Main.class.getResourceAsStream(path);
	else
	  {
	    try
	      {
		File file = new File(path);
		input = new FileInputStream(file);
	      }
	    catch(FileNotFoundException e)
	      {
		input = null;
	      }
	  }
	if (input == null)
	  {
	    printError("could not open file '" + path + "'");
	    return;
	  }	
	run(new InputStreamReader(input));
	input.close();
	if (!control.isSilent())
	  control.println("> file '" + path + "' read.");
      }
    catch(IOException e)
      {
	printError("error in reading file '" + path + "': " + e.getMessage());
      }
    catch(IllegalArgumentException e)
      {
	// internet explorer raises this on invalid file syntax
	printError("invalid file name '" + path + "'");
      }
  }

  // --------------------------------------------------------------------------
  // set option 'name', returns true if parsing was okay
  // --------------------------------------------------------------------------
  private boolean setOption(StreamTokenizer st) 
  {
    try
    {
      String name = parseString(st);
      if (name == null) return false;
      String fdef = parseString(st);
      if (fdef == null) return false;
      if (!fdef.equals(FDEF)) return false;
      
      // switch guard checking on/off
      if (name.equals("check"))
      {
	String value = parseString(st);
	if (value == null) return false;
	boolean set = value.equals("true");
	Model.setCheck(set);
	if (set)
	  control.println("> execution checks are on.");
	else
	  control.println("> execution checks are off.");
      }

      // switch guard checking on/off
      else if (name.equals("silent"))
      {
	String value = parseString(st);
	if (value == null) return false;
	boolean set = value.equals("true");
	control.setSilent(set);
      }

      // set universe of discourse
      else if (name.equals("universe"))
      {
	Term term = parseTerm(st);
	if (term == null) return false;
	try
	{
	  Value value = term.eval();
	  if (!(value instanceof Domain))
	    printError("value " + value + "is not a domain");
	  else
	  {
	    Model.setDomain((Domain)value);
	    control.println("> universe of discourse set.");
	  }
	}
	catch(EvalException e)
	{
	  // context is in an undefined state
	  Context.reset();
	  printError(e.getMessage());
	}
      }
      else
      {
	printError("unknown option '" + name +"'");
      }
    }
    catch(IOException e)
    {
      printError("IOException" + e.getMessage());
      control.abort();
    }
    return true;
  }

  // --------------------------------------------------------------------------
  // plot points denoted by 'value' in color 'color' on screen 'screen'
  // --------------------------------------------------------------------------
  private void plot(Value value, Color color, Screen screen) 
    throws EvalException
  {
    if (!(value instanceof Set))
      throw new EvalException("value is not a set: " + value.getString());
    Iterator iterator = ((Set)value).getIterator();
    while (iterator.hasNext())
    {
      Value v = iterator.next();
      if (!(v instanceof Tuple))
	throw new EvalException("value is not a tuple: " + v.getString());
      Tuple tuple = (Tuple)v;
      if (tuple.length() != 2)
	throw new EvalException
	  ("tuple is not of length 2: " + tuple.getString());
      Value value0 = tuple.elementAt(0);
      Value value1 = tuple.elementAt(1);
      if (!(value0 instanceof Nat))
	throw new EvalException
	  ("value is not a number: " + value0.getString());
      if (!(value1 instanceof Nat))
	throw new EvalException
	  ("value is not a number: " + value1.getString());
      Nat nat0 = (Nat)value0;
      Nat nat1 = (Nat)value1;
      screen.set(nat0.intValue(), nat1.intValue(), color);
    }
  }

  // --------------------------------------------------------------------------
  // plot relation (set of natural number pairs)
  // --------------------------------------------------------------------------
  private void plotRelation(Term term)
  {
    Screen screen = control.getScreen();
    if (screen == null)
    {
      printError("cannot plot without screen");
      return;
    }
    screen.clear();
    try
    {
      Value value = term.eval();
      plot(value, Color.blue, screen);
      control.println("> done.");
    }
    catch(EvalException e)
    {
      // context is in an undefined state
      Context.reset();
      printError(e.getMessage());
    }
    screen.flush();
  }

  // --------------------------------------------------------------------------
  // plot classes (set of relations)
  // --------------------------------------------------------------------------
  private void plotClasses(Term term)
  {
    Screen screen = control.getScreen();
    if (screen == null)
    {
      printError("cannot plot without screen");
      return;
    }
    Color colors[] = {Color.blue, Color.red, Color.green, Color.magenta, 
		      Color.yellow, Color.orange, Color.gray, Color.pink};
    screen.clear();
    try
    {
      Value value = term.eval();
      if (!(value instanceof Set))
	throw new EvalException("value is not a set: " + value.getString());
      Iterator iterator = ((Set)value).getIterator();
      int i = 0;
      while (iterator.hasNext())
      {
	plot(iterator.next(), colors[i], screen);
	i++;
	if (i == colors.length) i = 0;
      }
      control.println("> done.");
    }
    catch(EvalException e)
    {
      // context is in an undefined state
      Context.reset();
      printError(e.getMessage());
    }
    screen.flush();      
  }
}
// ----------------------------------------------------------------------------
// $Id: Interpreter.java,v 1.3 1999/07/12 13:07:16 schreine Exp $
// ----------------------------------------------------------------------------
