// ----------------------------------------------------------------------------
// $Id: Model.java,v 1.14 1999/07/12 13:07:16 schreine Exp $
// model holding predicate and function interpretations
//
// (c) 1999, Wolfgang Schreiner, see file COPYRIGHT
// http://www.risc.uni-linz.ac.at/software/formal
// ----------------------------------------------------------------------------
package Logic;
import Logic.Predicates.*;
import Logic.Functions.*;
import java.util.*;

public final class Model
{
  // we use public (not private) for member vars because of netscape bug

  // vectors of hashtables indexed by arity of predicate/function
  public Vector predicates = new Vector();
  public Vector functions = new Vector();

  // are guards to be checked?
  public boolean check = true;

  // --------------------------------------------------------------------------
  // construct model with builtin predicates and functions
  // --------------------------------------------------------------------------
  public Model()
  {
    // predicates
    _add(new True());
    _add(new False());
    _add(new Equals());
    _add(new In());
    _add(new IsSet());
    _add(new IsTuple());
    _add(new IsNat());
    _add(new LessEqual());

    // functions
    _add(new EmptySet());
    _add(new Join());
    _add(new Length());
    _add(new NaturalsFun());
    _add(new Add());
    _add(new Subtract());
    _add(new Multiply());
  }

  // the model used for static class references
  public static Model model = null;

  // --------------------------------------------------------------------------
  // set model used for static class references
  // --------------------------------------------------------------------------
  public static void setModel(Model _model)
  {
    model = _model;
  }

  // --------------------------------------------------------------------------
  // static class interface using default model
  // --------------------------------------------------------------------------
  public static void add(Predicate pred)
  {
    model._add(pred);
  }
  public static void add(Function fun)
  {
    model._add(fun);
  }
  public static void delete(Predicate pred)
  {
    model._delete(pred);
  }
  public static void delete(Function fun)
  {
    model._delete(fun);
  }
  public static Predicate getPredicate(String name, int arity)
  {
    return model._getPredicate(name, arity);
  }
  public static Function getFunction(String name, int arity)
  {
    return model._getFunction(name, arity);
  }
  public static void setDomain(Domain d)
  {
    model._setDomain(d);
  }
  public static Domain getDomain()
  {
    return model._getDomain();
  }
  public static boolean checkGuard()
  {
    return model._checkGuard();
  }
  public static void setCheck(boolean b)
  {
    model._setCheck(b);
  }

  // --------------------------------------------------------------------------
  // evaluate model to domain and return iterator
  // --------------------------------------------------------------------------
  public static Iterator iterator(Term term) throws EvalException
  {
    if (term == null)
    {
      Domain domain = getDomain();
      if (domain == null)
	throw new EvalException("no universe of discourse specified");
      return domain.getIterator();
    }
    Value value = term.eval();
    if (!(value instanceof Domain))
      throw new EvalException
	("value " + value.getString() + " does not denote a domain");
    return ((Domain)value).getIterator();
  }

  // --------------------------------------------------------------------------
  // add predicate 'pred' to model
  // --------------------------------------------------------------------------
  private void _add(Predicate pred)
  {
    int arity = pred.arity();
    if (predicates.size() <= arity)
      predicates.setSize(arity+1);
    Hashtable table = (Hashtable)predicates.elementAt(arity);
    if (table == null)
      {
	table = new Hashtable();
	predicates.setElementAt(table, arity);
      }
    table.put(pred.name(), pred);
  }

  // --------------------------------------------------------------------------
  // add function 'fun' to model
  // --------------------------------------------------------------------------
  private void _add(Function fun)
  {
    int arity = fun.arity();
    if (functions.size() <= arity)
      functions.setSize(arity+1);
    Hashtable table = (Hashtable)functions.elementAt(arity);
    if (table == null)
      {
	table = new Hashtable();
	functions.setElementAt(table, arity);
      }
    table.put(fun.name(), fun);
  }

  // --------------------------------------------------------------------------
  // delete predicate 'pred' from model
  // --------------------------------------------------------------------------
  private void _delete(Predicate pred)
  {
    int arity = pred.arity();
    Hashtable table = (Hashtable)predicates.elementAt(arity);
    table.remove(pred.name());
  }

  // --------------------------------------------------------------------------
  // delete function 'fun' from model
  // --------------------------------------------------------------------------
  private void _delete(Function fun)
  {
    int arity = fun.arity();
    Hashtable table = (Hashtable)functions.elementAt(arity);
    table.remove(fun.name());
  }

  // --------------------------------------------------------------------------
  // return predicate 'name/arity' from model; null if no such function
  // --------------------------------------------------------------------------
  private Predicate _getPredicate(String name, int arity)
  {
    if (predicates.size() <= arity) return null;
    Hashtable table = (Hashtable)predicates.elementAt(arity);
    if (table == null) return null;
    return (Predicate)table.get(name);
  }

  // --------------------------------------------------------------------------
  // return function 'name/arity' from model; null if no such function
  // --------------------------------------------------------------------------
  private Function _getFunction(String name, int arity)
  {
    if (functions.size() <= arity) return null;
    Hashtable table = (Hashtable)functions.elementAt(arity);
    if (table == null) return null;
    return (Function)table.get(name);
  }

  // our universe of discourse (only used for unrestricted quantifiers)
  protected Domain domain = null;

  // --------------------------------------------------------------------------
  // set universe of discourse
  // --------------------------------------------------------------------------
  private void _setDomain(Domain d)
  {
    domain = d;
  }

  // --------------------------------------------------------------------------
  // get universe of discourse
  // --------------------------------------------------------------------------
  private Domain _getDomain()
  {
    return domain;
  }

  // --------------------------------------------------------------------------
  // are guards to be checked?
  // --------------------------------------------------------------------------
  private boolean _checkGuard()
  {
    return check;
  }

  // --------------------------------------------------------------------------
  // set whether guards are to be checked
  // --------------------------------------------------------------------------
  private void _setCheck(boolean c)
  {
    check = c;
  }
}
// ----------------------------------------------------------------------------
// $Id: Model.java,v 1.14 1999/07/12 13:07:16 schreine Exp $
// ----------------------------------------------------------------------------


